@@ -4,6 +4,7 @@ local uri_patts = require "lpeg_patterns.uri"
 local basexx = require "basexx"
 local client = require "http.client"
 local new_headers = require "http.headers".new
+local http_cookie = require "http.cookie"
 local http_hsts = require "http.hsts"
 local http_socks = require "http.socks"
 local http_proxies = require "http.proxies"
@@ -15,10 +16,18 @@ local ce = require "cqueues.errno"
 local default_user_agent = string.format("%s/%s", http_version.name, http_version.version)
 local default_hsts_store = http_hsts.new_store()
 local default_proxies = http_proxies.new():update()
+local default_cookie_store = http_cookie.new_store()
+
+local default_h2_settings = {
+	ENABLE_PUSH = false;
+}
 
 local request_methods = {
 	hsts = default_hsts_store;
 	proxies = default_proxies;
+	cookie_store = default_cookie_store;
+	is_top_level = true;
+	site_for_cookies = nil;
 	expect_100_timeout = 1;
 	follow_redirects = true;
 	max_redirects = 5;
@@ -32,7 +41,7 @@ local request_mt = {
 }
 
 local EOF = lpeg.P(-1)
-local sts_patt = lpeg.Cf(lpeg.Ct(true) * http_patts.Strict_Transport_Security, rawset) * EOF
+local sts_patt = http_patts.Strict_Transport_Security * EOF
 local uri_patt = uri_patts.uri * EOF
 local uri_ref = uri_patts.uri_reference * EOF
 
@@ -44,7 +53,7 @@ local function new_from_uri(uri_t, headers)
 	end
 	local scheme = assert(uri_t.scheme, "URI missing scheme")
 	assert(scheme == "https" or scheme == "http" or scheme == "ws" or scheme == "wss", "scheme not valid")
-	local host = tostring(assert(uri_t.host, "URI must include a host")) -- tostring required to e.g. convert lpeg_patterns IPv6 objects
+	local host = assert(uri_t.host, "URI must include a host")
 	local port = uri_t.port or http_util.scheme_to_port[scheme]
 	local is_connect -- CONNECT requests are a bit special, see http2 spec section 8.3
 	if headers == nil then
@@ -77,8 +86,8 @@ local function new_from_uri(uri_t, headers)
 		else
 			field = "authorization"
 		end
-		local userinfo = http_util.decodeURIComponent(uri_t.userinfo) -- XXX: this doesn't seem right, but it's same behaviour as curl
-		headers:append(field, "basic " .. basexx.to_base64(userinfo), true)
+		local userinfo = http_util.decodeURIComponent(uri_t.userinfo) -- XXX: this doesn't seem right, but it's the same behaviour as curl
+		headers:upsert(field, "basic " .. basexx.to_base64(userinfo), true)
 	end
 	if not headers:has("user-agent") then
 		headers:append("user-agent", default_user_agent)
@@ -103,6 +112,7 @@ function request_methods:clone()
 	return setmetatable({
 		host = self.host;
 		port = self.port;
+		bind = self.bind;
 		tls = self.tls;
 		ctx = self.ctx;
 		sendname = self.sendname;
@@ -114,6 +124,9 @@ function request_methods:clone()
 
 		hsts = rawget(self, "hsts");
 		proxies = rawget(self, "proxies");
+		cookie_store = rawget(self, "cookie_store");
+		is_top_level = rawget(self, "is_top_level");
+		site_for_cookies = rawget(self, "site_for_cookies");
 		expect_100_timeout = rawget(self, "expect_100_timeout");
 		follow_redirects = rawget(self, "follow_redirects");
 		max_redirects = rawget(self, "max_redirects");
@@ -362,6 +375,18 @@ function request_methods:go(timeout)
 		end
 	end
 
+	if self.cookie_store then
+		local cookie_header = self.cookie_store:lookup_for_request(request_headers, host, self.site_for_cookies, self.is_top_level)
+		if cookie_header ~= "" then
+			if not cloned_headers then
+				request_headers = request_headers:clone()
+				cloned_headers = true
+			end
+			-- Append rather than upsert: user may have added their own cookies
+			request_headers:append("cookie", cookie_header)
+		end
+	end
+
 	local connection
 
 	local proxy = self.proxy
@@ -373,8 +398,16 @@ function request_methods:go(timeout)
 	if proxy then
 		if type(proxy) == "string" then
 			proxy = assert(uri_patt:match(proxy), "invalid proxy URI")
+			proxy.path = nil -- ignore proxy.path component
 		else
 			assert(type(proxy) == "table" and getmetatable(proxy) == nil and proxy.scheme, "invalid proxy URI")
+			proxy = {
+				scheme = proxy.scheme;
+				userinfo = proxy.userinfo;
+				host = proxy.host;
+				port = proxy.port;
+				-- ignore proxy.path component
+			}
 		end
 		if proxy.scheme == "http" or proxy.scheme == "https" then
 			if tls then
@@ -383,10 +416,8 @@ function request_methods:go(timeout)
 				local connect_request = new_connect(proxy, authority)
 				connect_request.proxy = false
 				connect_request.version = 1.1 -- TODO: CONNECT over HTTP/2
-				if tls then
-					if connect_request.tls then
-						error("NYI: TLS over TLS")
-					end
+				if connect_request.tls then
+					error("NYI: TLS over TLS")
 				end
 				-- Perform CONNECT request
 				local headers, stream, errno = connect_request:go(deadline and deadline-monotime())
@@ -404,10 +435,12 @@ function request_methods:go(timeout)
 				local sock = stream.connection:take_socket()
 				local err, errno2
 				connection, err, errno2 = client.negotiate(sock, {
+					host = host;
 					tls = tls;
 					ctx = self.ctx;
-					sendname = self.sendname ~= nil and self.sendname or host;
+					sendname = self.sendname;
 					version = self.version;
+					h2_settings = default_h2_settings;
 				}, deadline and deadline-monotime())
 				if connection == nil then
 					sock:close()
@@ -417,9 +450,6 @@ function request_methods:go(timeout)
 				if request_headers:get(":method") == "CONNECT" then
 					error("cannot use HTTP Proxy with CONNECT method")
 				end
-				if proxy.path ~= nil and proxy.path ~= "" then
-					error("an HTTP proxy cannot have a path component")
-				end
 				-- TODO: Check if :path already has authority?
 				local old_url = self:to_uri(false)
 				host = assert(proxy.host, "proxy is missing host")
@@ -446,6 +476,7 @@ function request_methods:go(timeout)
 				ctx = self.ctx;
 				sendname = self.sendname ~= nil and self.sendname or host;
 				version = self.version;
+				h2_settings = default_h2_settings;
 			}, deadline and deadline-monotime())
 			if connection == nil then
 				sock:close()
@@ -461,10 +492,12 @@ function request_methods:go(timeout)
 		connection, err, errno = client.connect({
 			host = host;
 			port = port;
+			bind = self.bind;
 			tls = tls;
 			ctx = self.ctx;
 			sendname = self.sendname;
 			version = self.version;
+			h2_settings = default_h2_settings;
 		}, deadline and deadline-monotime())
 		if connection == nil then
 			return nil, err, errno
@@ -580,6 +613,10 @@ function request_methods:go(timeout)
 		end
 	end
 
+	if self.cookie_store then
+		self.cookie_store:store_from_request(request_headers, headers, self.host, self.site_for_cookies)
+	end
+
 	if self.follow_redirects and headers:get(":status"):sub(1,1) == "3" then
 		stream:shutdown()
 		local new_req, err2, errno2 = self:handle_redirect(headers)
